#c #windows #linux 

Все дочерние потоки завершаются, когда завершился основной поток.

# Linux

## Библиотека

```c
#include <pthread.h>
```

**Как компилировать программу**

```bash
gcc source.c -pthread -o source
```

## Создание потока

В основном приводят в пример следующую функцию и её сигнатуру:

```c
int pthread_create(pthread_t *thread, const pthread_attr_t *attr, void *(*start)(void *), void *arg);
```

Происходит вызов потока и программа идёт дальше.

Как и в Windows, поток начинается с выполнения функции, которая указана аргументом **start** , затем передается аргумент **arg**, который напрямую транслируется в start.

Разберёмся:

* **start** - указатель на функцию, которая принимает аргумент
* **arg** - аргумент функции.

Хоть мы и передаём, void* это не значит, что мы транслируем ничего. Как раз наоборот в функцию мы транслируем, что угодно т.к. **void* - указатель**.

```c
#include <pthread.h>
#include <stdio.h>
#include <unistd.h>

static void * threadFunc(void *arg)
{   
    printf("%s",(char *) arg); // привидение типа
    return (void*) 0;
}

int main()
{
    pthread_t t1;
    
    int s = pthread_create(&t1, NULL, threadFunc, "Hello world\n");
    sleep(1);//задержка для того, чтобы функция потока гарантированно выполнилась
    return 0;
}
```

## Завершение потока

* return
* exit() - завершение всех потоков, в том числе и главного.
* pthread_exit() - завершает вызываемый поток и указывает возвращаемое значение.

Функция pthread_join() может взаимодействовать с потоком, через pthread_exit().

Если главный поток вызывает pthread_exit() (вместо return или exit), остальные потоки продолжат выполнение.

```c
pthread_exit(void* retval)
```

Аргумент **retval**, хранит возвращаемое потомком значение. При этом retval, должен быть вне стека потока, т.к. его значение в стеке будет удалено вместе со стеком, при завершении потока.

## Идентификатор потока

Каждый поток внутри процесса имеет свой уникальный идентификатор.

Его можно получить из вызывающего потока следующей функцией:

```c
pthread_create();
```

Также с помощью другой функции получить собственный идентификатор:

```c
pthread_t pthread_self(void);
```

Проверка тождественность двух идентификаторов:

```c
int pthread_equal(pthread_t t1, pthread_t t2);
```

**Возвращает не  нуль если равны, иначе нуль.**

## Присоединение к завершенному потоку

Функция, ждёт завершение потока обозначенного аргументом thread, (если поток завершён, то сразу возвращается):

```c
int pthread_join(pthread_t thread, void** retval);
```

**Возвращает - при успешном завршении или положительное число, если ошибка**

Аргумент этой функции, является ненулевым указателем, функция получит копию возвращаемого значение завершенного потока, то есть значение, указанное при выполнении потоком инструкции return.

```c
#include <pthread.h>
#include <stdio.h>
#include <unistd.h>

static void * threadFunc(void *arg)
{   
    printf("%s",(char *) arg);
    sleep(3);//теперь делаем задержку в потоке, чтобы проверить механизм ожидания его завершения в вызывающем потоке.
    return (void*) "Я завершился";
 }

int main()
{
    pthread_t t1;
    int s = pthread_create(&t1, NULL, threadFunc, "Hello world\n");
    void* res;
    s = pthread_join(t1,&res);
    printf("Завершенный поток сказал: %s",(char*)res);
    return 0;
}
```

## Отсоединение от потока

Иногда нам не важно значение, которое вернул поток. Нам нужно чтобы система автоматически освободила ресурсы и удалила поток, когда тот завершится.

Мы можем это отметить через следующую функцию:

```c
int pthread_detach(ptthread_t thread);
```

**Вернёт - при успешном завершении или положительное число, если возникла ошибка.**

К примеру мы можем отсоединить поток, чтобы он не был подвержен функции exit:

```c
pthread_detach(pthread_self());
```

Отсоединившейся поток:

* Не зависит от return в главном процессе (потоке).
* Не зависит от exit().

## Синхронизация потока

Хоть и потоки могут общаться и делится информацией, как через глобальные переменные так и через каналы. Нам нужно регулировать их доступ к данным.

```c
#include <pthread.h>
#include <stdio.h>
#include <unistd.h>

int counter=0;//глобальная переменная, значение которой мы увеличиваем в нескольких потоках

static void * threadFunc(void *arg)
{   
    for (size_t i = 0; i < *(int*)arg; i++)
    {
        counter++;
    }   
}

int main()
{
    pthread_t t1,t2,t3;

    int n = 100500;//количество итераций в цикле

    pthread_create(&t1, NULL, threadFunc, &n);
    pthread_create(&t2, NULL, threadFunc, &n);
    pthread_create(&t3, NULL, threadFunc, &n);   
    
    pthread_join(t1,NULL);
    pthread_join(t2,NULL);
    pthread_join(t3,NULL);

    printf("Результат инкрементации: %d",counter); //по идее counter должен был получиться n*3, но на практике это не так

    return 0;
}
```

Для того чтобы разграничить доступ есть mutex (mutual exclusion - взаимное исключение), которое гарантирует что только один поток имеет доступ к переменной в определённый промежуток времени.

Они позволяют обеспечить атомарный доступ к разделяемому ресурсу или защита общих переменных.

Два состояния мьютекса:

* Закрытое (блокирование);
* Открытое (разблокированное).

**В одно время один поток блокирует мьютексом один ресурс.** 

Действие с ресурсом:

* Закрытие мьютексом чего-либо, назначает владельца ресурса.
* Владелец ресурсом - тот кто его блокировал.
* Только владелец открывает ресурс заново.
* Попытка закрыть ресурс либо отклоняется, либо выдаёт ошибку.

**Мьютекс выделен в виде статической переменной или в создан динамически при выполнении программы.**

```c
pthread_mutex_t; // Переменная мьютекса

pthread_mutex_t mtx = PTHREAD_MUTEX_INITIALIZER; // Статически
```

Пример:

```c
#include <pthread.h>
#include <stdio.h>
#include <unistd.h>

pthread_mutex_t mtx = PTHREAD_MUTEX_INITIALIZER;//создаем и инициализируем мьютекс

int counter=0;//глобальная переменная, значение которой мы увеличиваем в нескольких потоках

static void * threadFunc(void *arg)
{   
    for (size_t i = 0; i < *(int*)arg; i++)
    {
        pthread_mutex_lock(&mtx);
        counter++;
        pthread_mutex_unlock(&mtx);
    }   
}

int main()
{
    pthread_t t1,t2,t3;

    int n = 100500;//количество итераций в цикле

    pthread_create(&t1, NULL, threadFunc, &n);
    pthread_create(&t2, NULL, threadFunc, &n);
    pthread_create(&t3, NULL, threadFunc, &n);   

    pthread_join(t1,NULL);
    pthread_join(t2,NULL);
    pthread_join(t3,NULL);

    printf("Результат инкрементации: %d",counter); //теперь counter = n*3

    return 0;
}
```

# Windows

## Создание потоков

```c
CreateThread (  
NULL  /*указатель на структуру безопасности (NULL по умолчанию)*/,
0 /*размер стека (0 по умолчанию)*/, 
func /*указатель на функцию, которая будет выполняться в потоке*/,
NULL /*указатель на аргумент функции потока (NULL - пустой указатель, без аргументов)*/,
0 /*флаги создания потока. 0 - по умолчанию.*/,
0 /*ID потока (0 - автоматически).*/);
```

Функция потока имеет следующую **строгую** сигнатуру:

```c
DWORD WINAPI func(LPVOID param);
```

## Ожидание и завершение потока

Ожидание выполнение потоков:

```c
WaitForMultipleObjects
(count /*Количество потоков, завершения которых необходимо ждать*/, 
hF /*Указатель на дескрипторы (Массив дескрипторов)*/,
TRUE /*Флаг ожидания. Истина - будет ждать завершения всех потоков из count, Ложь - ожидание Завершения одного любого потока.*/, 
INFINITE /*Время ожидания завершения потока.*/);
```

Выход из потока:

```c  
ExitThread(0);
```

## Синхронизация потока
 
**Если мы ставим ресурс, на пауза два или более раз, то нам нужно снять его столько же раз с паузы.**

### Критическая секция

**Критическая секция** - метод, который фиксирует выполнение кожа только одним потоком.

* Создаётся соответствующая переменная
* До начала выполнения потока необходима инициализировать критическую секцию с помощью соответствующей функции
* Участок кода, который контролируется критической секцией должен находится между инструкциями.
* После окончания работы потоков можно освободить критическую секцию с помощью функции.

```c
#define _CRT_SECURE_NO_WARNINGS
#include <Windows.h>
#include <stdio.h>
#include <malloc.h>
#define I_WILL_WAIT 10

CRITICAL_SECTION section = { 0 }; //Критическая секция

VOID Crit(VOID);
DWORD WINAPI Thread1(DWORD param);
DWORD WINAPI Thread2(DWORD param);

int count = 0;

HANDLE h[3];

VOID Crit(VOID)
{
   DWORD tmp1 = 1;
   DWORD tmp2 = 2;
   DWORD tmp3 = 3;

   InitializeCriticalSection(&section);

   h[0] = CreateThread(NULL, 0, Thread2, tmp1, 0, 0);       
   h[1] = CreateThread(NULL, 0, Thread2, tmp2, 0, 0);          
   h[2] = CreateThread(NULL, 0, Thread2, tmp3, 0, 0);

   WaitForMultipleObjects(3, h, TRUE, INFINITE);

   DeleteCriticalSection(&section);

   printf("count = %d\n", count);
}

DWORD WINAPI Thread1(DWORD param)
{
	for (int i=0;i<10;i++)
	{
		Sleep(I_WILL_WAIT);
		count++;        
		printf("count = %d, potok = %d\n", count, param);
	}

	ExitThread(0);
}

DWORD WINAPI Thread2(DWORD param)
{
	EnterCriticalSection(&section);
	
	for (int i = 0; i < 10; i++)
	{
		Sleep(I_WILL_WAIT);                                   
		count++;                                            
		printf("count = %d, potok = %d\n", count, param);
	}
	
	LeaveCriticalSection(&section);
	ExitThread(0);
}
```



